import logging

# Things created in the context.
from rockingester_api.collectors.collectors import (
    Collectors,
    rockingester_collectors_set_default,
)

# Base class.
from rockingester_api.context_base import ContextBase

logger = logging.getLogger(__name__)


class Context(ContextBase):
    """
    Client context for a rockingester_collector object.
    On entering, it creates the object according to the specification (a dict).
    On exiting, it closes client connection.

    The aenter and aexit methods are exposed for use by an enclosing context.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification):
        self.__specification = specification

    # ----------------------------------------------------------------------------------------
    async def aenter(self):
        """ """

        # Build the object according to the specification.
        self.interface = Collectors().build_object(self.__specification)

        # If there is more than one collector, the last one defined will be the default.
        rockingester_collectors_set_default(self.interface)

    # ----------------------------------------------------------------------------------------
    async def aexit(self):
        """ """

        if self.interface is not None:
            await self.interface.close_client_session()

            # Clear the global variable.  Important between pytests.
            rockingester_collectors_set_default(None)



import asyncio
import logging
import time
from pathlib import Path

# Things xchembku provides.
from xchembku_api.datafaces.context import Context as XchembkuDatafaceClientContext
from xchembku_api.datafaces.datafaces import xchembku_datafaces_get_default
from xchembku_api.models.crystal_plate_model import CrystalPlateModel

# Client context creator.
from rockingester_api.collectors.context import Context as CollectorClientContext

# Server context creator.
from rockingester_lib.collectors.context import Context as CollectorServerContext

# Base class for the tester.
from tests.base import Base

logger = logging.getLogger(__name__)


# ----------------------------------------------------------------------------------------
class TestCollectorDirectPoll:
    """
    Test collector interface by direct call.
    """

    def test(self, constants, logging_setup, output_directory):

        # Configuration file to use.
        configuration_file = "tests/configurations/direct_poll.yaml"

        CollectorTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class TestCollectorService:
    """
    Test collector interface through network interface.
    """

    def test(self, constants, logging_setup, output_directory):

        # Configuration file to use.
        configuration_file = "tests/configurations/service.yaml"

        CollectorTester().main(constants, configuration_file, output_directory)


# ----------------------------------------------------------------------------------------
class CollectorTester(Base):
    """
    Test scraper collector's ability to automatically discover files and push them to xchembku.
    """

    # ----------------------------------------------------------------------------------------
    async def _main_coroutine(self, constants, output_directory):
        """
        This tests the collector behavior.

        First it starts the xchembku and collector and loads a single plate barcode into the database.

        Then, while the collector thread is running, it creates some scrapable images.

        The test then waits a while for the scraping to be done, and verifies the outputs.

        These are in 3 barcodes.  The first matches the barcode in the database, so it gets scraped.
        The second matches no barcode in the database, so it is moved to the nobarcode area and not added to the database.
        The third barcode is not scraped because it is not configured in the ingest_only_barcodes list.
        """

        # Get the multiconf from the testing configuration yaml.
        multiconf = self.get_multiconf()

        # Load the multiconf into a dict.
        multiconf_dict = await multiconf.load()

        # Reference the dict entry for the xchembku dataface.
        xchembku_dataface_specification = multiconf_dict[
            "xchembku_dataface_specification"
        ]

        # Make the xchembku client context, expected to be direct (no server).
        xchembku_client_context = XchembkuDatafaceClientContext(
            xchembku_dataface_specification
        )

        collector_specification = multiconf_dict["rockingester_collector_specification"]
        # Make the server context.
        collector_server_context = CollectorServerContext(collector_specification)

        # Make the client context.
        collector_client_context = CollectorClientContext(collector_specification)

        # Remember the collector specification so we can assert some things later.
        self.__ingested_directory = Path(multiconf_dict["ingested_directory"])
        self.__nobarcode_directory = Path(multiconf_dict["nobarcode_directory"])

        image_count = 2

        # Start the client context for the direct access to the xchembku.
        async with xchembku_client_context:
            # Start the collector client context.
            async with collector_client_context:
                # And the collector server context which starts the coro.
                async with collector_server_context:
                    await self.__run_part1(image_count, constants, output_directory)

                logger.debug(
                    "------------ restarting collector server --------------------"
                )

                # Start the server again.
                # This covers the case where collector starts by finding existing entries in the database and doesn't double-collect those on disk.
                async with collector_server_context:
                    await self.__run_part2(image_count, constants, output_directory)

    # ----------------------------------------------------------------------------------------

    async def __run_part1(self, image_count, constants, output_directory):
        """ """
        # Reference the xchembku object which the context has set up as the default.
        xchembku = xchembku_datafaces_get_default()

        # Make the plate on which the wells reside.
        visit = "cm00001-1"
        crystal_plate_model = CrystalPlateModel(
            formulatrix__plate__id=10,
            barcode="98ab",
            visit=visit,
        )

        await xchembku.upsert_crystal_plates([crystal_plate_model])

        # Get list of images before we create any of the scrape-able files.
        records = await xchembku.fetch_crystal_wells_filenames()

        assert len(records) == 0, "images before any new files are put in scrapable"

        plates_directory = Path(output_directory) / "SubwellImages"

        # Make the scrapable directory with some files.
        plate_directory1 = plates_directory / "98ab_2023-04-06_RI1000-0276-3drop"
        plate_directory1.mkdir(parents=True)
        for i in range(10, 10 + image_count):
            filename = plate_directory1 / ("98ab_%02dA_1.jpg" % (i))
            with open(filename, "w") as stream:
                stream.write("")

        # Make another scrapable directory with a different barcode.
        plate_directory2 = plates_directory / "98ac_2023-04-06_RI1000-0276-3drop"
        plate_directory2.mkdir(parents=True)
        for i in range(10, 10 + image_count + 1):
            filename = plate_directory2 / ("98ac_%02dA_1.jpg" % (i))
            with open(filename, "w") as stream:
                stream.write("")

        # Make yet another scrapable directory with a different barcode.
        plate_directory3 = plates_directory / "98ad_2023-04-06_RI1000-0276-3drop"
        plate_directory3.mkdir(parents=True)
        for i in range(10, 10 + image_count + 2):
            filename = plate_directory3 / ("98ad_%02dA_1.jpg" % (i))
            with open(filename, "w") as stream:
                stream.write("")

        # Wait for all the images to appear.
        time0 = time.time()
        timeout = 5.0
        while True:

            # Get all images.
            records = await xchembku.fetch_crystal_wells_filenames()

            # Stop looping when we got the images we expect.
            if len(records) >= image_count:
                break

            if time.time() - time0 > timeout:
                raise RuntimeError(
                    f"only {len(records)} images out of {image_count}"
                    f" registered within {timeout} seconds"
                )
            await asyncio.sleep(1.0)

        # Wait a couple more seconds to make sure there are no extra images appearing.
        await asyncio.sleep(2.0)

        # Get all images in the database.
        records = await xchembku.fetch_crystal_wells_filenames()
        assert len(records) == image_count, "images after scraping"

        # The two plate directories should have been emptied.
        count = sum(1 for _ in plate_directory1.glob("*") if _.is_file())
        assert count == 0, "first plate_directory"

        count = sum(1 for _ in plate_directory2.glob("*") if _.is_file())
        assert count == 0, "second plate_directory"

        # The third plate directory is left intact.
        count = sum(1 for _ in plate_directory3.glob("*") if _.is_file())
        assert count == image_count + 2, "third plate_directory"

        # We should have ingested the first barcode.
        count = sum(1 for _ in self.__ingested_directory.glob("*") if _.is_dir())
        assert count == 1, f"ingested_directory {str(self.__ingested_directory)}"
        count = sum(
            1
            for _ in (self.__ingested_directory / plate_directory1.name).glob("*")
            if _.is_file()
        )
        assert (
            count == image_count
        ), f"ingested_directory images {str(self.__ingested_directory)}"

        # We should have send the second barcode to the nobarcode area.
        count = sum(1 for _ in self.__nobarcode_directory.glob("*") if _.is_dir())
        assert count == 1, f"nobarcode_directory {str(self.__nobarcode_directory)}"
        count = sum(
            1
            for _ in (self.__nobarcode_directory / plate_directory2.name).glob("*")
            if _.is_file()
        )
        assert (
            count == image_count + 1
        ), f"nobarcode_directory images {str(self.__nobarcode_directory)}"

    # ----------------------------------------------------------------------------------------

    async def __run_part2(self, image_count, constants, output_directory):
        """ """
        # Reference the xchembku object which the context has set up as the default.
        xchembku = xchembku_datafaces_get_default()

        await asyncio.sleep(2.0)
        # Get all images after servers start up and run briefly.
        records = await xchembku.fetch_crystal_wells_filenames()

        assert len(records) == image_count, "images after restarting scraper"



class Keywords:
    COMMAND = "collectors::keywords::command"
    PAYLOAD = "collectors::keywords::payload"


class Commands:
    EXECUTE = "collectors::commands::execute"


class Types:
    AIOHTTP = "rockingester_lib.collectors.aiohttp"
    DIRECT_POLL = "rockingester_lib.collectors.direct_poll"



import logging
import multiprocessing
import threading

# Utilities.
from dls_utilpack.callsign import callsign
from dls_utilpack.require import require

# Base class which maps flask tasks to methods.
from dls_utilpack.thing import Thing

# Base class for an aiohttp server.
from rockingester_lib.base_aiohttp import BaseAiohttp

# Factory to make a Collector.
from rockingester_lib.collectors.collectors import Collectors

# Collector protocolj things.
from rockingester_lib.collectors.constants import Commands, Keywords

logger = logging.getLogger(__name__)

thing_type = "rockingester_lib.collectors.aiohttp"


# ------------------------------------------------------------------------------------------
class Aiohttp(Thing, BaseAiohttp):
    """
    Object representing an image collector.
    The behavior is to start a direct collector coro task or process
    to waken every few seconds and scan for incoming files.

    Then start up a web server to handle generic commands and queries.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification=None, predefined_uuid=None):
        Thing.__init__(self, thing_type, specification, predefined_uuid=predefined_uuid)
        BaseAiohttp.__init__(
            self, specification["type_specific_tbd"]["aiohttp_specification"]
        )

        self.__direct_collector = None

    # ----------------------------------------------------------------------------------------
    def callsign(self) -> str:
        """
        Put the class name into the base class's call sign.

        Returns:
            str: call sign withi class name in it
        """

        return "%s %s" % ("Collector.Aiohttp", BaseAiohttp.callsign(self))

    # ----------------------------------------------------------------------------------------
    def activate_process(self) -> None:
        """
        Activate the direct collector and web server in a new process.

        Meant to be called from inside a newly started process.
        """

        try:
            multiprocessing.current_process().name = "collector"

            self.activate_process_base()

        except Exception as exception:
            logger.exception("exception in collector process", exc_info=exception)

        logger.debug(f"[PIDAL] {callsign(self)} is returning from activate_process")

    # ----------------------------------------------------------------------------------------
    def activate_thread(self, loop) -> None:
        """
        Activate the direct collector and web server in a new thread.

        Meant to be called from inside a newly created thread.
        """

        try:
            threading.current_thread().name = "collector"

            self.activate_thread_base(loop)

        except Exception as exception:
            logger.exception(
                f"unable to start {callsign(self)} thread", exc_info=exception
            )

    # ----------------------------------------------------------------------------------------
    async def activate_coro(self) -> None:
        """
        Activate the direct collector and web server in a two asyncio tasks.
        """

        try:
            # Build a local collector for our back-end.
            self.__direct_collector = Collectors().build_object(
                self.specification()["type_specific_tbd"][
                    "direct_collector_specification"
                ]
            )

            logger.info("[COLSHUT] calling self.__direct_collector.activate()")
            # Get the local implementation started.
            await self.__direct_collector.activate()

            # ----------------------------------------------
            logger.info("[COLSHUT] calling BaseAiohttp.activate_coro_base(self)")
            await BaseAiohttp.activate_coro_base(self)

            logger.info("[COLSHUT] returning")

        except Exception as exception:
            raise RuntimeError(
                "exception while starting collector server"
            ) from exception

    # ----------------------------------------------------------------------------------------
    async def direct_shutdown(self) -> None:
        """
        Shut down any started sub-process or coro tasks.

        Then call the base_direct_shutdown to shut down the webserver.

        """
        logger.info(
            f"[COLSHUT] in direct_shutdown self.__direct_collector is {self.__direct_collector}"
        )

        # ----------------------------------------------
        if self.__direct_collector is not None:
            # Disconnect our local dataface connection, i.e. the one which holds the database connection.
            logger.info("[COLSHUT] awaiting self.__direct_collector.deactivate()")
            await self.__direct_collector.deactivate()
            logger.info(
                "[COLSHUT] got return from self.__direct_collector.deactivate()"
            )

        # ----------------------------------------------
        # Let the base class stop the server listener.
        await self.base_direct_shutdown()

    # ----------------------------------------------------------------------------------------
    async def __do_locally(self, function, args, kwargs):
        """"""

        # logger.info(describe("function", function))
        # logger.info(describe("args", args))
        # logger.info(describe("kwargs", kwargs))

        function = getattr(self.__direct_collector, function)

        response = await function(*args, **kwargs)

        return response

    # ----------------------------------------------------------------------------------------
    async def dispatch(self, request_dict, opaque):
        """"""

        command = require("request json", request_dict, Keywords.COMMAND)

        if command == Commands.EXECUTE:
            payload = require("request json", request_dict, Keywords.PAYLOAD)
            response = await self.__do_locally(
                payload["function"], payload["args"], payload["kwargs"]
            )
        else:
            raise RuntimeError("invalid command %s" % (command))

        return response



import logging
from typing import Dict

# Things created in the context.
from rockingester_lib.collectors.collectors import Collectors, collectors_set_default

# Base class for an asyncio context.
from rockingester_lib.contexts.base import Base as ContextBase

logger = logging.getLogger(__name__)


thing_type = "rockingester_lib.collectors.context"


class Context(ContextBase):
    """
    Asyncio context for a collector object.
    On entering, it creates the object according to the specification (a dict).
    If specified, it starts the server as a coroutine, thread or process.
    If not a server, then it will instatiate a direct access to a collector.
    On exiting, it commands the server to shut down and/or releases the direct access resources.
    """

    # ----------------------------------------------------------------------------------------
    def __init__(self, specification: Dict):
        """
        Constructor.

        Args:
            specification (Dict): specification of the collector object to be constructed within the context.
                The only key in the specification that relates to the context is "start_as", which can be "coro", "thread", "process" or None.
                All other keys in the specification relate to creating the collector object.
        """
        ContextBase.__init__(self, thing_type, specification)

    # ----------------------------------------------------------------------------------------
    async def aenter(self) -> None:
        """
        Asyncio context entry.

        Starts and activates service as specified.

        Establishes the global (singleton-like) default collector.
        """

        # Build the object according to the specification.
        self.server = Collectors().build_object(self.specification())

        # If there is more than one collector, the last one defined will be the default.
        collectors_set_default(self.server)

        if self.context_specification.get("start_as") == "coro":
            await self.server.activate_coro()

        elif self.context_specification.get("start_as") == "thread":
            await self.server.start_thread()

        elif self.context_specification.get("start_as") == "process":
            await self.server.start_process()

        # Not running as a service?
        elif self.context_specification.get("start_as") == "direct":
            # We need to activate the tick() task.
            await self.server.activate()

    # ----------------------------------------------------------------------------------------
    async def aexit(self) -> None:
        """
        Asyncio context exit.

        Stop service if one was started and releases any client resources.
        """

        if self.server is not None:
            if self.context_specification.get("start_as") == "process":
                logger.info(
                    "[NEWSHUT] in context exit, sending shutdown to client process"
                )
                # Put in request to shutdown the server.
                await self.server.client_shutdown()
                logger.info(
                    "[NEWSHUT] in context exit, sent shutdown to client process"
                )

            if self.context_specification.get("start_as") == "coro":
                await self.server.direct_shutdown()

            if self.context_specification.get("start_as") == "direct":
                await self.server.deactivate()
